/*
 *  Modified by bertel@post4.tele.dk, to be used with Wain.
 *
 *  Copyright (c) 1986 by University of Toronto.
 *  Written by Henry Spencer.  Not derived from licensed software.
 *
 *  Permission is granted to anyone to use this software for any
 *  purpose on any computer system, and to redistribute it freely,
 *  subject to the following restrictions:
 *
 *  1. The authors is not responsible for the consequences of use of
 *     this software, no matter how awful, even if they arise
 *     from defects in it.
 *
 *  2. The origin of this software must not be misrepresented, either
 *     by explicit claim or by omission.
 *
 *  3. Altered versions must be plainly marked as such, and must not
 *     be misrepresented as being the original software.
 *
 * Beware that some of this code is subtly aware of the way operator
 * precedence is structured in regular expressions.  Serious changes in
 * regular-expression syntax might require a total rethink.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "regexp.h"

/*
 * The "internal use only" fields in regexp.h are present to pass info from
 * compile to execute that permits the execute phase to run lots faster on
 * simple cases.  They are:
 *
 * regstart  char that must begin a match; '\0' if none obvious
 * reganch   is the match anchored (at beginning-of-line only)?
 * regmust   string (pointer into program) that match must include, or NULL
 * regmlen   length of regmust string
 *
 * Regstart and reganch permit very fast decisions on suitable starting points
 * for a match, cutting down the work a lot.  Regmust permits fast rejection
 * of lines that cannot possibly match.  The regmust tests are costly enough
 * that regcomp() supplies a regmust only if the r.e. contains something
 * potentially expensive (at present, the only such thing detected is * or +
 * at the start of the r.e., which can involve a lot of backup).  Regmlen is
 * supplied because the test in regexec() needs it and regcomp() is computing
 * it anyway.
 */

/*
 * Structure for regexp "program".  This is essentially a linear encoding
 * of a nondeterministic finite-state machine (aka syntax charts or
 * "railroad normal form" in parsing technology).  Each node is an opcode
 * plus a "next" pointer, possibly plus an operand.  "Next" pointers of
 * all nodes except BRANCH implement concatenation; a "next" pointer with
 * a BRANCH on both ends of it is connecting two alternatives.  (Here we
 * have one of the subtle syntax dependencies:  an individual BRANCH (as
 * opposed to a collection of them) is never concatenated with anything
 * because of operator precedence.)  The operand of some types of node is
 * a literal string; for others, it is a node leading into a sub-FSM.  In
 * particular, the operand of a BRANCH node is the first node of the branch.
 * (NB this is *not* a tree structure:  the tail of the branch connects
 * to the thing following the set of BRANCHes.)  The opcodes are:
 */

/* definition    opnd? meaning */
enum
{
  END,        /* no    End of program. */
  BOL,        /* no    Match "" at beginning of line. */
  EOL,        /* no    Match "" at end of line. */
  ANY,        /* no    Match any one character. */
  ANYOF,      /* str   Match any character in this string. */
  ANYBUT,     /* str   Match any character not in this string. */
  BRANCH,     /* node  Match this alternative, or the next... */
  BACK,       /* no    Match "", "next" ptr points backward. */
  EXACTLY,    /* str   Match this string. */
  NOTHING,    /* no    Match empty string. */
  STAR,       /* node  Match this (simple) thing 0 or more times. */
  PLUS,       /* node  Match this (simple) thing 1 or more times. */
  OPEN =  20, /* no    Mark this point in input as start of #n. */
              /*       OPEN + 1 is number 1, etc. */
  CLOSE = 30  /* no    Analogous to OPEN. */
};

/*
 * Opcode notes:
 *
 * BRANCH  The set of branches constituting a single choice are hooked
 *         together with their "next" pointers, since precedence prevents
 *         anything being concatenated to any individual branch.  The
 *         "next" pointer of the last BRANCH in a choice points to the
 *         thing following the whole choice.  This is also where the
 *         final "next" pointer of each individual branch points; each
 *         branch starts with the operand node of a BRANCH node.
 *
 * BACK    Normal "next" pointers all implicitly point forward; BACK
 *         exists to make loop structures possible.
 *
 * STAR,PLUS  '?', and complex '*' and '+', are implemented as circular
 *         BRANCH structures using BACK.  Simple cases (one character
 *         per match) are implemented with STAR and PLUS for speed
 *         and to minimize recursive plunges.
 *
 * OPEN,CLOSE  ...are numbered at compile time.
 */

/*
 * A node is one char of opcode followed by two chars of "next" pointer.
 * "Next" pointers are stored as two 8-bit pieces, high order first.  The
 * value is a positive offset from the opcode of the node containing it.
 * An operand, if any, simply follows the node.  (Note that much of the
 * code generation knows about this implicit relationship.)
 *
 * Using two bytes for the "next" pointer is vast overkill for most things,
 * but allows patterns to get big without disasters.
 */
#define OP(p)  (*(p))
#define NEXT(p)  (((*((p)+1)& 0xFF )<<8) + (*((p)+2) & 0xFF))
#define OPERAND(p)  ((p) + 3)

#define MAGIC 0234
/*
 * Utility definitions.
 */
#ifndef CHARBITS
#define  UCHARAT(p)  ((int)*(unsigned char *)(p))
#else
#define  UCHARAT(p)  ((int)*(p)&CHARBITS)
#endif

#define ISMULT(c)  ((c) == '*' || (c) == '+' || (c) == '?')
#define META  "^$.[()|?+*\\"

/*
 * Flags to be passed up and down.
 */
#define  HASWIDTH  01  /* Known never to match null string. */
#define  SIMPLE    02  /* Simple enough to be STAR/PLUS operand. */
#define  SPSTART   04  /* Starts with * or +. */
#define  WORST      0  /* Worst case. */


/*
 - compile() - compile a regular expression into internal code
 *
 * We can't allocate space until we know how big the compiled form will be,
 * but we can't compile it (and thus know how big it is) until we've got a
 * place to put the code.  So we cheat:  we compile it twice, once with code
 * generation turned off and size counting turned on, and once "for real".
 * This also means that we don't allocate space until we are sure that the
 * thing really will compile successfully, and we never have to move the
 * code and thus invalidate pointers into it.  (Note that it has to be in
 * one piece because free() must be able to free it all.)
 *
 * Beware that the optimization-preparation code in here knows about some
 * of the structure of the compiled regexp.
 */

struct regexp_type
{
  char regstart;
  char reganch;
  char *regmust;
  int regmlen;
  char program[1];           /* The actual size of this is allocated on the fly */
};

BOOL reg_exp_class::compile(const char *exp)
{
  char *scan;
  char *longest;
  unsigned int len;
  int flags;

  if(prog)
    free(prog);

  if(exp == NULL)
  {
    regerror("NULL argument");
    return FALSE;
  }
  /* First pass: determine size, legality. */
  regparse = exp;
  regnpar = 1;
  regsize = 0L;
  regcode = &regdummy;
  regc((char )MAGIC);
  if(reg(0, &flags) == NULL)
    return FALSE;

  /* Small enough for pointer-storage convention? */
  if(regsize >= 0x10000)
  {
    regerror("regexp too big");
    return FALSE;
  }
  /* Allocate space. */
  prog = (regexp_type *)malloc(sizeof(regexp_type) + (unsigned )regsize);

  if(prog == NULL)
  {
    regerror("out of space");
    return FALSE;
  }

  /* Second pass: emit code. */
  regparse = exp;
  regnpar = 1;
  regcode = prog->program;
  regc((char )MAGIC);
  if(reg(0, &flags) == NULL)
  {
    free(prog);
    prog = NULL;
    return FALSE;
  }

  /* Dig out information for optimizations. */
  prog->regstart = '\0';  /* Worst-case defaults. */
  prog->reganch = 0;
  prog->regmust = NULL;
  prog->regmlen = 0;
  scan = prog->program + 1;      /* First BRANCH. */

  if(OP(regnext(scan)) == END)
  {    /* Only one top-level choice. */
    scan = OPERAND(scan);

    /* Starting-point info. */
    if(OP(scan) == EXACTLY)
      prog->regstart = *OPERAND(scan);
    else if(OP(scan) == BOL)
      prog->reganch++;

    /*
     * If there's something expensive in the r.e., find the
     * longest literal string that must appear and make it the
     * regmust.  Resolve ties in favor of later strings, since
     * the regstart check works with the beginning of the r.e.
     * and avoiding duplication strengthens checking.  Not a
     * strong reason, but sufficient in the absence of others.
     */
    if(flags & SPSTART)
    {
      longest = NULL;
      len = 0;
      for(; scan != NULL; scan = regnext(scan))
        if(OP(scan) == EXACTLY && strlen(OPERAND(scan)) >= len)
        {
          longest = OPERAND(scan);
          len = strlen(OPERAND(scan));
        }
      prog->regmust = longest;
      prog->regmlen = len;
    }
  }
  return TRUE;
}

/*
 - reg - regular expression, i.e. main body or parenthesized thing
 *
 * Caller must absorb opening parenthesis.
 *
 * Combining parenthesis handling with the base level of regular expression
 * is a trifle forced, but the need to tie the tails of the branches to what
 * follows makes it hard to avoid.
 */
char *reg_exp_class::reg(int paren, int *flagp)
{
  char *ret;
  char *br;
  char *ender;
  int parno = regnpar;
  int flags;

  *flagp = HASWIDTH;  /* Tentatively. */

  /* Make an OPEN node, if parenthesized. */
  if(paren)
  {
    if(regnpar >= NSUBEXP)
    {
      regerror("too many ()");
      return NULL;
    }
    parno = regnpar;
    regnpar++;
    ret = regnode(OPEN + parno);
  }
  else
    ret = NULL;

  /* Pick up the branches, linking them together. */
  br = regbranch(&flags);
  if(br == NULL)
    return NULL;
  if(ret != NULL)
    regtail(ret, br);  /* OPEN -> first. */
  else
    ret = br;

  if(!(flags & HASWIDTH))
    *flagp &= ~HASWIDTH;

  *flagp |= flags & SPSTART;

  while (*regparse == '|')
  {
    regparse++;
    br = regbranch(&flags);
    if(br == NULL)
      return NULL;
    regtail(ret, br);  /* BRANCH -> BRANCH. */
    if(!(flags & HASWIDTH))
      *flagp &= ~HASWIDTH;
    *flagp |= flags & SPSTART;
  }

  /* Make a closing node, and hook it on the end. */
  ender = regnode((paren) ? CLOSE + parno : END);
  regtail(ret, ender);

  /* Hook the tails of the branches to the closing node. */
  for(br = ret; br != NULL; br = regnext(br))
    regoptail(br, ender);

  /* Check for proper termination. */
  if(paren && *regparse++ != ')')
  {
    regerror("unmatched ()");
    return NULL;
  }
  else if(!paren && *regparse != '\0')
  {
    if(*regparse == ')')
    {
      regerror("unmatched ()");
      return NULL;
    }
    else
    {
      regerror("junk on end");  /* "Can't happen". */
      return NULL;
    }
    /* NOTREACHED */
  }

  return ret;
}

/*
 - regbranch - one alternative of an | operator
 *
 * Implements the concatenation operator.
 */
char *reg_exp_class::regbranch(int *flagp)
{
  char *ret;
  char *chain;
  char *latest;
  int flags;

  *flagp = WORST;    /* Tentatively. */

  ret = regnode(BRANCH);
  chain = NULL;
  while(*regparse != '\0' && *regparse != '|' && *regparse != ')')
  {
    latest = regpiece(&flags);
    if(latest == NULL)
      return NULL;
    *flagp |= flags & HASWIDTH;
    if(chain == NULL)  /* First piece. */
      *flagp |= flags & SPSTART;
    else
      regtail(chain, latest);
    chain = latest;
  }
  if(chain == NULL)  /* Loop ran zero times. */
    regnode(NOTHING);

  return ret;
}

/*
 - regpiece - something followed by possible [*+?]
 *
 * Note that the branching code sequences used for ? and the general cases
 * of * and + are somewhat optimized:  they use the same NOTHING node as
 * both the endmarker for their branch list and the body of the last branch.
 * It might seem that this node could be dispensed with entirely, but the
 * endmarker role is not redundant.
 */
char *reg_exp_class::regpiece(int *flagp)
{
  char *ret;
  char op;
  char *next;
  int flags;

  ret = regatom(&flags);

  if(ret == NULL)
    return NULL;

  op = *regparse;
  if(!ISMULT(op))
  {
    *flagp = flags;
    return ret;
  }

  if(!(flags & HASWIDTH) && op != '?')
  {
    regerror("*+ operand could be empty");
    return NULL;
  }

  *flagp = (op != '+') ? (WORST | SPSTART) : (WORST | HASWIDTH);

  if(op == '*' && (flags & SIMPLE))
    reginsert(STAR, ret);
  else if(op == '*')
  { /* Emit x* as (x&|), where & means "self". */
    reginsert(BRANCH, ret);          /* Either x */
    regoptail(ret, regnode(BACK));   /* and loop */
    regoptail(ret, ret);             /* back */
    regtail(ret, regnode(BRANCH));   /* or */
    regtail(ret, regnode(NOTHING));  /* null. */
  }
  else if(op == '+' && (flags & SIMPLE))
    reginsert(PLUS, ret);
  else if(op == '+')
  {
    /* Emit x+ as x(&|), where & means "self". */
    next = regnode(BRANCH);           /* Either */
    regtail(ret, next);
    regtail(regnode(BACK), ret);      /* loop back */
    regtail(next, regnode(BRANCH));   /* or */
    regtail(ret, regnode(NOTHING));   /* null. */
  }
  else if(op == '?')
  {
    /* Emit x? as (x|) */
    reginsert(BRANCH, ret);          /* Either x */
    regtail(ret, regnode(BRANCH));   /* or */
    next = regnode(NOTHING);         /* null. */
    regtail(ret, next);
    regoptail(ret, next);
  }
  regparse++;
  if(ISMULT(*regparse))
  {
    regerror("nested *?+");
    return NULL;
  }

  return ret;
}

/*
 - regatom - the lowest level
 *
 * Optimization:  gobbles an entire sequence of ordinary characters so that
 * it can turn them into a single node, which is smaller to store and
 * faster to run.  Backslashed characters are exceptions, each becoming a
 * separate node; the code is simpler that way and it's not worth fixing.
 */
char *reg_exp_class::regatom(int *flagp)
{
  char *ret;
  int flags;

  *flagp = WORST;    /* Tentatively. */

  switch (*regparse++)
  {
    case '^':
      ret = regnode(BOL);
      break;
    case '$':
      ret = regnode(EOL);
      break;
    case '.':
      ret = regnode(ANY);
      *flagp |= HASWIDTH | SIMPLE;
      break;
    case '[':
      {
        int class_t;
        int classend;

        if(*regparse == '^')
        {  /* Complement of range. */
          ret = regnode(ANYBUT);
          regparse++;
        }
        else if(*regparse == '[')
        {
          regerror("Unexpected '[' in []");
          return NULL;
        }
        else
          ret = regnode(ANYOF);

        if(*regparse == ']' || *regparse == '-')
          regc(*regparse++);
        while(*regparse != '\0' && *regparse != ']')
        {
          if(*regparse == '-')
          {
            regparse++;
            if(*regparse == ']' || *regparse == '\0')
              regc('-');
            else
            {
              class_t = UCHARAT(regparse-2)+1;
              classend = UCHARAT(regparse);
              if(class_t > classend + 1)
              {
                regerror("invalid [] range");
                return NULL;
              }
              for(; class_t <= classend; class_t++)
                regc(class_t);
              regparse++;
            }
          }
          else if(*regparse == '\\')
          {
            regparse++;
            switch(*regparse)
            {
            case '\\':
            case ']':
            case '-':
               regc(*regparse);
               regparse++;
               break;
            case 't':
               regc('\t');
               regparse++;
               break;
            default:
               regerror("Invalid esc. seq.");
               return 0;
               break;
            }
          }
          else
            regc(*regparse++);
        }
        regc('\0');
        if(*regparse != ']')
        {
          regerror("unmatched []");
          return NULL;
        }
        regparse++;
        *flagp |= HASWIDTH | SIMPLE;
      }
      break;
    case '(':
      ret = reg(1, &flags);
      if(ret == NULL)
        return(NULL);
      *flagp |= flags&(HASWIDTH|SPSTART);
      break;
    case '\0':
    case '|':
    case ')':
      regerror("internal urp");  /* Supposed to be caught earlier. */
      return NULL;
    case '?':
    case '+':
    case '*':
      regerror("?+* follows nothing");
      return NULL;
    case '\\':
      if(*regparse == '\0')
      {
        regerror("trailing \\");
        return NULL;
      }
      ret = regnode(EXACTLY);
      regc(*regparse++);
      regc('\0');
      *flagp |= HASWIDTH|SIMPLE;
      break;
    default:
      {
        int len;
        char ender;

        regparse--;
        len = strcspn(regparse, META);
        if(len <= 0)
        {
          regerror("internal disaster");
          return NULL;
        }
        ender = *(regparse + len);
        if(len > 1 && ISMULT(ender))
          len--;    /* Back off clear of ?+* operand. */
        *flagp |= HASWIDTH;
        if(len == 1)
          *flagp |= SIMPLE;
        ret = regnode(EXACTLY);
        while(len > 0)
        {
          regc(*regparse++);
          len--;
        }
        regc('\0');
      }
      break;
  }

  return ret;
}

/*
 - regnode - emit a node
 */
 /* Location. */
char *reg_exp_class::regnode(char op)
{
  char *ret;
  char *ptr;

  ret = regcode;
  if(ret == &regdummy)
  {
    regsize += 3;
    return ret;
  }

  ptr = ret;
  *ptr++ = op;
  *ptr++ = '\0';    /* Null "next" pointer. */
  *ptr++ = '\0';
  regcode = ptr;

  return ret;
}

/*
 - regc - emit (if appropriate) a byte of code
 */
void reg_exp_class::regc(char b)
{
  if(regcode != &regdummy)
    *regcode++ = b;
  else
    regsize++;
}

/*
 - reginsert - insert an operator in front of already-emitted operand
 *
 * Means relocating the operand.
 */
void reg_exp_class::reginsert(char op, char *opnd)
{
  char *src;
  char *dst;
  char *place;

  if(regcode == &regdummy)
  {
    regsize += 3;
    return;
  }

  src = regcode;
  regcode += 3;
  dst = regcode;
  while (src > opnd)
    *--dst = *--src;

  place = opnd;    /* Op node, where operand used to be. */
  *place++ = op;
  *place++ = '\0';
  *place = '\0';
}

/*
 - regtail - set the next-pointer at the end of a node chain
 */
void reg_exp_class::regtail(char *p, char *val)
{
  char *scan;
  char *temp;
  int offset;

  if(p == &regdummy)
    return;

  /* Find last node. */
  scan = p;
  for(;;)
  {
    temp = regnext(scan);
    if(temp == NULL)
      break;
    scan = temp;
  }

  if(OP(scan) == BACK)
    offset = scan - val;
  else
    offset = val - scan;
  *(scan+1) = (offset >> 8) & 0xFF;
  *(scan+2) = offset & 0xFF;
}

/*
 - regoptail - regtail on operand of first argument; nop if operandless
 */
void reg_exp_class::regoptail(char *p, char *val)
{
  /* "Operandless" and "op != BRANCH" are synonymous in practice. */
  if(p == NULL || p == &regdummy || OP(p) != BRANCH)
    return;
  regtail(OPERAND(p), val);
}

/*
 - find() - match a regexp against a string
 */

const char *reg_exp_class::find(const char *string, int *len)
{
  const char *s;
  const char *r;

  /* Be paranoid... */
  if(prog == NULL || string == NULL)
  {
    regerror("NULL parameter");
    return NULL;
  }

  /* Check validity of program. */
  if(UCHARAT(prog->program) != MAGIC)
  {
    regerror("corrupted program");
    return NULL;
  }
  reg_len = 0;
  max_len = 0;
  /* If there is a "must appear" string, look for it. */
  if(prog->regmust != NULL)
  {
    s = string;
    while ((s = strchr(s, prog->regmust[0])) != NULL)
    {
      if(strncmp(s, prog->regmust, prog->regmlen) == 0)
        break;  /* Found it. */
      s++;
    }
    if(s == NULL)  /* Not present. */
    {
      if(len)
        *len = 0;
      return NULL;
    }
  }

  /* Mark beginning of line for ^ . */
  regbol = string;

  /* Simplest case:  anchored match need be tried only once. */
  if(prog->reganch)
  {
    r = regtry(string);
    if(reg_len > max_len)
      max_len = reg_len;
    if(len)
      *len = max_len;
    return r;
  }

  /* Messy cases:  unanchored match. */
  s = string;
  if(prog->regstart != '\0')
  { /* We know what char it must start with. */
    while((s = strchr(s, prog->regstart)) != NULL)
    {
      reg_len = 0;
      max_len = 0;
      if((r = regtry(s)) != NULL)
      {
        if(reg_len > max_len)
          max_len = reg_len;
        if(len)
          *len = max_len;
        return r;
      }
      s++;
    }
  }
  else
  { /* We don't -- general case. */
    do
    {
      reg_len = 0;
      max_len = 0;
      if((r = regtry(s)) != NULL)
      {
        if(reg_len > max_len)
          max_len = reg_len;
        if(len)
          *len = max_len;
        return r;
      }
    }
    while (*s++ != '\0');
  }

  /* Failure. */
  if(len)
    *len = 0;
  return NULL;
}

/* regtry - try match at specific point */

const char *reg_exp_class::regtry(const char *string)
{
  const char *r;

  reginput = string;
  if((r = regmatch(prog->program + 1, 0)) != NULL)
  {
    return r;
  }
  else
    return NULL;
}

/*
 - regmatch - main matching routine
 *
 * Conceptually the strategy is simple:  check to see whether the current
 * node matches, call self recursively to see whether the rest matches,
 * and then act accordingly.  In practice we make some effort to avoid
 * recursion, in particular by going through "ordinary" nodes (that don't
 * need to know whether the rest of the match failed) by a loop instead of
 * by recursion.
 */
const char *reg_exp_class::regmatch(char *prog, int nest)
{
  char *scan;    /* Current node. */
  char *next;    /* Next node. */
  const char *r_save = reginput;

  scan = prog;

  while(scan != NULL)
  {
    next = regnext(scan);

    switch (OP(scan))
    {
      case BOL:
        if(reginput != regbol)
          return NULL;
        break;
      case EOL:
        if(*reginput != '\0')
          return NULL;
        break;
      case ANY:
        if(*reginput == '\0')
          return NULL;
        reginput++;
        reg_len++;
        break;
      case EXACTLY:
        {
          int len;
          char *opnd;

          opnd = OPERAND(scan);
          /* Inline the first character, for speed. */
          if(*opnd != *reginput)
            return NULL;
          len = strlen(opnd);
          if(len > 1 && strncmp(opnd, reginput, len) != 0)
            return NULL;
          reginput += len;
          reg_len += len;
        }
        break;
      case ANYOF:
        if(*reginput == '\0' || strchr(OPERAND(scan), *reginput) == NULL)
          return NULL;
        reginput++;
        reg_len++;
        break;
      case ANYBUT:
        if(*reginput == '\0' || strchr(OPERAND(scan), *reginput) != NULL)
          return NULL;
        reginput++;
        reg_len++;
        break;
      case NOTHING:
        break;
      case BACK:
        break;
      case OPEN + 1:
      case OPEN + 2:
      case OPEN + 3:
      case OPEN + 4:
      case OPEN + 5:
      case OPEN + 6:
      case OPEN + 7:
      case OPEN + 8:
      case OPEN + 9:
        if(regmatch(next, nest + 1))
        {
          /*
           * Don't set startp if some later
           * invocation of the same parentheses
           * already has.
           */
          return r_save;
        }
        else
          return NULL;
      case CLOSE + 1:
      case CLOSE + 2:
      case CLOSE + 3:
      case CLOSE + 4:
      case CLOSE + 5:
      case CLOSE + 6:
      case CLOSE + 7:
      case CLOSE + 8:
      case CLOSE + 9:
        {
          if(regmatch(next, nest + 1))
          {
            /*
             * Don't set endp if some later
             * invocation of the same parentheses
             * already has.
             */
            return r_save;
          }
          else
          {
            return NULL;
          }
        }
      case BRANCH:
        {
          const char *save;
          if(OP(next) != BRANCH)    /* No choice. */
            next = OPERAND(scan);  /* Avoid recursion. */
          else
          {
            do
            {
              save = reginput;
              if(regmatch(OPERAND(scan), nest + 1))
                return r_save;
              reginput = save;
              scan = regnext(scan);
              if(!nest)
              {
                if(reg_len > max_len)
                  max_len = reg_len;
                reg_len = 0;
              }
            }
            while (scan != NULL && OP(scan) == BRANCH);
            return NULL;
            /* NOTREACHED */
          }
        }
        break;
      case STAR:
      case PLUS:
        {
          char nextch;
          int no;
          const char *save;
          int min;

          /*
           * Lookahead to avoid useless match attempts
           * when we know what character comes next.
           */
          nextch = '\0';
          if(OP(next) == EXACTLY)
            nextch = *OPERAND(next);
          min = (OP(scan) == STAR) ? 0 : 1;
          save = reginput;
          no = regrepeat(OPERAND(scan));
          reg_len += no;
          while (no >= min)
          {
            /* If it could work, try it. */
            if(nextch == '\0' || *reginput == nextch)
              if(regmatch(next, nest + 1))
              {
                return r_save;
              }
            /* Couldn't or didn't -- back up. */
            no--;
            reginput = save + no;
          }
          return NULL;
        }
      case END:
        /* Success! */
        return r_save;
      default:
        regerror("memory corruption");
        return NULL;
    }

    scan = next;
  }

  /*
   * We get here only if there's trouble -- normally "case END" is
   * the terminating point.
   */
  regerror("corrupted pointers");
  return NULL;
}

/*
 - regrepeat - repeatedly match something simple, report how many
 */
int reg_exp_class::regrepeat(char *p)
{
  int count = 0;
  const char *scan;
  char *opnd;

  scan = reginput;
  opnd = OPERAND(p);
  switch (OP(p))
  {
    case ANY:
      count = strlen(scan);
      scan += count;
      break;
    case EXACTLY:
      while (*opnd == *scan)
      {
        count++;
        scan++;
      }
      break;
    case ANYOF:
      while (*scan != '\0' && strchr(opnd, *scan) != NULL)
      {
        count++;
        scan++;
      }
      break;
    case ANYBUT:
      while (*scan != '\0' && strchr(opnd, *scan) == NULL)
      {
        count++;
        scan++;
      }
      break;
    default:    /* Oh dear.  Called inappropriately. */
      regerror("internal foulup");
      count = 0;  /* Best compromise. */
      break;
  }
  reginput = scan;

  return count;
}

/*
 - regnext - dig the "next" pointer out of a node
 */
char *reg_exp_class::regnext(char *p)
{
  int offset;

  if(p == &regdummy)
    return NULL;

  offset = NEXT(p);

  if(offset == 0)
    return NULL;

  if(OP(p) == BACK)
    return p - offset;
  else
    return p + offset;
}

void reg_exp_class::regerror(const char *msg)
{
  strncpy(err_msg, msg, MAX_ERR_MSG_SIZE);
}

reg_exp_class::reg_exp_class()
{
  prog = NULL;
  err_msg[0] = 0;
}

reg_exp_class::~reg_exp_class()
{
  if(prog)
    free(prog);
}

const char *reg_exp_class::get_last_error(void)
{
  return err_msg;
}

void reg_exp_class::clear_error(void)
{
  err_msg[0] = 0;
}

